;; Generated package description from beframe.el  -*- mode: lisp-data; no-byte-compile: t; lexical-binding:t -*-
(define-package "beframe" "1.5.0.0.20260828.18" "Isolate buffers per frame" '((emacs "28.1")) :commit "f1824ddbda0c6fb09745e4404e7521a42e2fb258" :authors '(("Protesilaos" . "info@protesilaos.com")) :maintainer '("Protesilaos" . "info@protesilaos.com") :url "https://github.com/protesilaos/beframe")
